/*
    bpcfg
    Bayesian PCFG induction.
    $Id: bpcfg.cpp,v 1.25 2015/03/10 01:24:28 daichi Exp $
    
 */
#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <vector>
#include <string>
#include "pcfg.h"
#include "pcfgs.h"
#include "pcfgt.h"
#include "util.h"
#define  K_DEFAULT	3
#define	 ITER_DEFAULT	1
#define	 GAMMA_DEFAULT	0.1
#define	 ETA_DEFAULT	0.01
#define	 ALPHA_DEFAULT	1.0
#define	 BETA_DEFAULT	4.0

using namespace std;

static char opts[] = "a:b:g:e:K:N:STh";

void
usage ()
{
	printf("bpcfg, Bayesian PCFG for unsupervised data.\n");
	printf("usage: %% %s OPTIONS train model\n", "bpcfg");
	printf("OPTIONS\n"
	       " -K states  number of hidden categories (default %d)\n"
	       " -N iter    number of MCMC iterations (default %d)\n"
	       " -g gamma   Dirichlet prior for state transitions (default %lg)\n"
	       " -e eta     Dirichlet prior for lexical emissions (default %lg)\n"
	       " -a alpha   Beta prior for random slices (default %lg)\n"
	       " -b beta    Beta prior for random slices (default %lg)\n",
	       K_DEFAULT, ITER_DEFAULT,
	       GAMMA_DEFAULT, ETA_DEFAULT, ALPHA_DEFAULT, BETA_DEFAULT);
	printf("$Id: bpcfg.cpp,v 1.25 2015/03/10 01:24:28 daichi Exp $\n");
	exit(0);
}

int
main (int argc, char *argv[])
{
	vector<vector<int> > data;
	char *train = NULL, *model = NULL;
	double gamma = GAMMA_DEFAULT;
	double eta = ETA_DEFAULT;
	double alpha = ALPHA_DEFAULT;
	double beta  = BETA_DEFAULT; 
	int K = K_DEFAULT, iter = ITER_DEFAULT;
	int c;
#if defined SLICE_T
	PCFGT *pcfg;
#elif defined SLICE_S
	PCFGS *pcfg;
#else
	PCFG *pcfg;
#endif

	while ((c = getopt(argc, argv, opts)) != -1)
	{
		switch (c) {
		case 'g' : gamma = atof(optarg); break;
		case 'e' : eta = atof(optarg); break;
		case 'a' : alpha = atof(optarg); break;
		case 'b' : beta = atof(optarg); break;
		case 'K' : K = atoi(optarg); break;
		case 'N' : iter = atoi(optarg); break;
		case 'h' : usage (); break;
		default  : usage (); break;
		}
	}
	if (argc - optind != 2)
		usage ();
	else
	{
#if defined SLICE_T
		pcfg = new PCFGT (K+1, gamma, eta, alpha, beta);
#elif defined SLICE_S
		pcfg = new PCFGS (K+1, gamma, eta);
#else
		pcfg = new PCFG (K+1, gamma, eta);
#endif
		train = argv[optind];
		model = argv[optind + 1];
	}
	printf("iter = %d, train = %s, model = %s\n", iter, train, model);

	/* read data */
	if (!pcfg->read(train, data))
	{
		fprintf(stderr, "reading data error.\n");
		exit(1);
	}

	/* train */
	if (!pcfg->train(data, iter))
	{
		error("training failed.");
		exit (1);
	} else {
		pcfg->save (model);
	}

	return 0;
}

