/*
    pcfg.cpp
    Bayesian PCFG implementation.
    $Id: pcfg.cpp,v 1.44 2015/01/28 02:46:40 daichi Exp $

*/
#include <cstdio>
#include <cmath>
#include <vector>
#include <string>
#include <iostream>
#include <fstream>
#include <algorithm>
#include "pcfg.h"
#include "tree.h"
#include "random.h"
#include "util.h"
#define	 INTERVAL	100

using namespace std;

Tree *
PCFG::parse (vector<int> &s)
{
	compute_cky (s);
	return sample (s, 0, 0, s.size() - 1);
	
}

double
PCFG::expand (int x, int y, int z)
{
	return grammar->expand (x, y, z);
}

double
PCFG::expand (int x, int w)
{
	return grammar->expand (x, w);
}
	
Tree *
PCFG::sample (vector<int> &s, int x, int i, int k)
{
	static vector<double> q;
	static vector<pair<int,pair<int,int> > > t;
	int j, y, z, l, L = k - i;
	double p;

	if (i == k)
	{
		return new Tree (x, s[i]);
	}

	/* allocate */
	q.reserve (L * K * K);
	t.reserve (L * K * K);
	q.clear ();
	t.clear ();

	/* sample */
	for (j = 0; j < L; j++)
	{
		for (y = 1; y < K; y++)
		{
			for (z = 1; z < K; z++)
			{
				p = log (expand(x, y, z)) +
					 alpha[i][i+j][y] + alpha[i+j+1][k][z];
				q.push_back (p);
				t.push_back (cons(j, cons(y, z)));
			}
		}
	}

	if ((l = lmultinom(q)) < 0) {
		error("invalid draw!");
		return NULL;
	} else {
		j = t[l].first;
		y = t[l].second.first;
		z = t[l].second.second;
		
		return new Tree (x, sample (s, y, i, i + j),
				    sample (s, z, i + j + 1, k));
	}
	
}

void
PCFG::compute_cky (vector<int> &s)
{
	int i, j, k, n, L = s.size();
	int x, y, z;

	/* allocate */
	if (L > alpha.size()) {
		alpha.resize (L);
		for (i = 0; i < L; i++) {
			alpha[i].resize (L);
			for (j = 0; j < L; j++) {
				alpha[i][j].resize (K);
			}
		}
	}

	/* initialize */
	for (i = 0; i < L; i++) {
		for (k = 1; k < K; k++) {
			alpha[i][i][k] = log (expand(k, s[i]));
		}
	}

	/* iterate */
	
	for (n = 1; n < L; n++)
	{
		for (i = 0; i < L - n; i++)
		{
			if (n == L - 1) {
				alpha[i][i+n][0] = cky_sum (0, i, n);
			}
			else {
				for (x = 1; x < K; x++)
				{
					alpha[i][i+n][x] = cky_sum (x, i, n);
				}
			}
		}
	}

}

double
PCFG::cky_sum (int x, int i, int n)
{
	int j, y, z;
	double l, lik = 0;

	for (y = 1; y < K; y++)
	{
		for (z = 1; z < K; z++)
		{
			l = 0;
			for (j = 0; j < n; j++)
			{
				l = logsumexp (l, alpha[i][i+j][y] +
					          alpha[i+j+1][i+n][z], (j == 0));
			}
			lik = logsumexp (lik, l + log(expand(x, y, z)));
		}
	}

	return lik;
}

bool
PCFG::add_counts (Tree *tree)
{
	return grammar->add_counts (tree);
}

bool
PCFG::remove_counts (Tree *tree)
{
	return grammar->remove_counts (tree);
}

double
PCFG::treelik (Tree *tree)
{
	return grammar->treelik (tree);
}

double
PCFG::polyalik (Tree *tree)
{
	return grammar->polyalik (tree);
}
    
bool
PCFG::train (vector<vector<int> > &data, int iter)
{
	static vector<int> seq;
	int i, j, n, N = data.size();
	double s, z, lik, newlik, diff;
	Tree *newtree;
	int accept;

	/* init */
	seq.resize (N);
	tree.resize (N);
	grammar->V = ntypes (data);
	for (i = 0; i < N; i++)
		seq[i] = i;
	
	fprintf(stderr, "N = %d, K = %d, V = %d, gamma = %lg, eta = %lg\n",
		N, K, grammar->V, grammar->gamma, grammar->eta);

	for (j = 0; j < iter; j++)
	{
		z = 0; s = 0; accept = 0;
		random_shuffle (seq.begin(), seq.end(), irand);
		
		for (i = 0; i < N; i++)
		{
			if ((i + 1) % INTERVAL == 0) {
				fprintf(stderr, "parsing %4d/%d..\r", i+1, N);
				fflush(stderr);
			}
			n = seq[i];
			
			if (j > 0) {
				remove_counts (tree[n]);
			}


			/* sample */
			newtree = parse (data[n]);
			newlik = polyalik (newtree);
			
			if (j == 0)
				diff = 0;
			else
			{
				lik = polyalik (tree[n]);
				diff = newlik - lik
					+ treelik(tree[n]) - treelik(newtree);
			}
			
			if (bernoulli(exp(diff)))  /* accept */
			{
				if (j > 0) delete (tree[n]);
				tree[n] = newtree;
				lik = newlik;
				accept++;
			}
			
			add_counts (tree[n]);
				
			z += lik;
			s += data[n].size();
			
		}
		printf("iter[%2d] : PPL = %4.2f (accept=%.2f%%)\n",
		       j + 1, exp(- z / s), (double) accept / N * 100);
	}
	    
	return true;
}

bool
PCFG::save (char *model)
{
	save_tree (model);
	save_dict (model);
	save_grammar (model);
}

bool
PCFG::save_dict (char *model)
{
	dict->save (model);
}

bool
PCFG::save_grammar (char *model)
{
	grammar->save (model);
}

bool
PCFG::save_tree (char *model)
{
	int i, N = tree.size();
	ofstream ofs (strconcat(model, ".tree"));

	if (!ofs) {
		perror("ofstream");
		return false;
	}

	for (i = 0; i < N; i++)
	{
		ofs << tree[i]->stringify(dict) << endl;
	}

	ofs.close();

	return true;
}

bool
PCFG::write (char *model)
{
	save_tree (model);
	write_grammar (model);
}

bool
PCFG::write_grammar (char *model)
{
	int x, y, z, v;
	FILE *fp;

	if ((fp = fopen(strconcat(model, ".grammar"), "w")) == NULL)
	{
		perror("fopen");
		return false;
	}

	for (x = 0; x < K; x++)
	{
		for (y = 1; y < K; y++)
		{
			for (z = 1; z < K; z++)
			{
				fprintf(fp, "%.7e\t%d\t%d\t%d\n",
					expand(x, y, z), x, y, z);
			}
		}
	}
	for (x = 1; x < K; x++)
	{
		for (v = 1; v <= grammar->V; v++)
		{
			fprintf(fp, "%.7e\t%d\t%s\n",
				expand(x, v), x, dict->word(v).c_str());
		}
	}

	fclose(fp);
	
}

bool
PCFG::read (char *file, vector<vector<int> > &data)
{
	vector<string> toks;
	string line;
	int i, w;
	ifstream ifs (file);

	if (!ifs) {
		perror("ifstream");
		return false;
	}

	while (getline(ifs, line, '\n'))
	{
		toks = parse_line (line);

		if (toks.size() > 0)
		{
			vector<int> row;
			
			for (i = 0; i < toks.size(); i++)
			{
				row.push_back (dict->id (toks[i]));
			}
			data.push_back (row);
		}
	}

	return true;
}
	
