/*
    pcfgs.cpp
    Bayesian PCFG with slice sampler.
    $Id: pcfgs.cpp,v 1.15 2015/03/05 11:53:28 daichi Exp $

*/
#include <cstdio>
#include <cstdlib>
#include <vector>
#include <string>
#include <iostream>
#include <fstream>
#include <algorithm>
#include <cmath>
#include "pcfgs.h"
#include "tree.h"
#include "random.h"
#include "util.h"
#define	 INTERVAL	100

using namespace std;

typedef struct {
	vector<double> *slice;
	Grammar *grammar;
} args;

Tree *
PCFGS::parse (vector<int> &s, Tree *tree, vector<double> &slice)
{
	set_slice (tree, slice);
	compute_cky (s, slice);
	return sample (s, 0, 0, s.size() - 1, slice);
}

Tree *
PCFGS::sample (vector<int> &s, int x, int i, int k, vector<double> &slice)
{
	static vector<double> q;
	static vector<pair<int,pair<int,int> > > t;
	int j, a, y, z, L = k - i;
	double l;

	if (i == k)
	{
		return new Tree (x, s[i]);
	}

	/* allocate */
	q.reserve (L * K * K);
	t.reserve (L * K * K);
	q.clear ();
	t.clear ();

	/* sample */
	for (j = 0; j < L; j++)
	{
		for (y = 1; y < K; y++)
		{
			for (z = 1; z < K; z++)
			{
				if (isinf(alpha[i][i+j][y]) ||
				    isinf(alpha[i+j+1][k][z]))
					;
				else {
					l = log (expand(x, y, z)) +
						alpha[i][i+j][y] + alpha[i+j+1][k][z];
					q.push_back (l);
					t.push_back (cons(j, cons(y, z)));
				}
			}
		}
	}

	if ((a = lmultinom(q)) < 0) {
		error("invalid draw!");
		exit(1);
		return NULL;
	} else {
		j = t[a].first;
		y = t[a].second.first;
		z = t[a].second.second;

		return new Tree (x, sample (s, y, i, i + j, slice),
				    sample (s, z, i + j + 1, k, slice));
	}
	
}

void
PCFGS::compute_cky (vector<int> &s, vector<double> &slice)
{
	int i, j, k, n, x, L = s.size();

	/* allocate */
	if (L > alpha.size()) {
		alpha.resize (L);
		for (i = 0; i < L; i++) {
			alpha[i].resize (L);
			for (j = 0; j < L; j++) {
				alpha[i][j].resize (K);
			}
		}
	}

	/* initialize */
	for (i = 0; i < L; i++) {
		for (k = 1; k < K; k++) {
			alpha[i][i][k] = log (expand(k, s[i]));
		}
		for (j = i+1; j < L; j++) {
			for (k = 0; k < K; k++) {
				alpha[i][j][k] = 0;
			}
		}
	}

	/* iterate */
	
	for (n = 1; n < L; n++)
	{
		for (i = 0; i < L - n; i++)
		{
			if (n == L - 1) {
				alpha[i][i+n][0] = cky_sum (0, i, n, slice);
			}
			else {
				for (x = 1; x < K; x++)
				{
					alpha[i][i+n][x] = cky_sum (x, i, n, slice);
				}
			}
		}
	}

#if 0
	/* print */
	for (k = 0; k < K; k++)
	{
		printf("[%d]:\n", k);
		for (i = 0; i < L; i++)
		{
			for (j = 0; j < L; j++)
			{
				if (i <= j)
					printf("%.4f ", alpha[i][j][k]);
			}
			printf("\n");
		}
	}
#endif

}

double
PCFGS::cky_sum (int x, int i, int n, vector<double> &slice)
{
	int j, y, z;
	double p, l, lik = -inf;

	for (y = 1; y < K; y++)
	{
		for (z = 1; z < K; z++)
		{
			l = -inf;
			p = expand (x, y, z);

			for (j = 0; j < n; j++)
			{
				if (p > slice[i+j])
				{
					l = logsumexp (l, alpha[i][i+j][y] +
						          alpha[i+j+1][i+n][z], (j == 0));
				}
			}
			lik = logsumexp (lik, l);
		}
	}

	return lik;
}

void
set_slice_func (int n, int x, int y, int z, void *varg)
{
 	args *arg = (args *)varg;

 	(*arg->slice)[n - 1] =
		RANDOM * arg->grammar->expand (x, y, z);
}

void
PCFGS::set_slice (Tree *tree, vector<double> &slice)
{
	args arg;
	arg.slice = &slice;
	arg.grammar = grammar;

	if (tree) {
		tree->splits (set_slice_func, (void *)&arg);
	}
}

double
PCFGS::expand (int x, int y, int z)
{
	return grammar->expand (x, y, z);
}

double
PCFGS::expand (int x, int w)
{
	return grammar->expand (x, w);
}
	
bool
PCFGS::add_counts (Tree *tree)
{
	return grammar->add_counts (tree);
}

bool
PCFGS::remove_counts (Tree *tree)
{
	return grammar->remove_counts (tree);
}

double
PCFGS::treelik (Tree *tree)
{
	return grammar->treelik (tree);
}
    
double
PCFGS::polyalik (Tree *tree)
{
	return grammar->polyalik (tree);
}

bool
PCFGS::train (vector<vector<int> > &data, int iter)
{
	static vector<int> seq;
	int i, j, n, N = data.size();
	double s, z, lik, newlik, diff;
	Tree *newtree;
	int accept;

	/* init */
	seq.resize (N);
	tree.resize (N);
	slices.resize (N);
	for (n = 0; n < N; n++)
	{
		seq[n] = n;
		tree[n] = NULL;
		slices[n].resize (data[n].size() - 1);
		for (i = 0; i < slices[n].size(); i++)
		{
			slices[n][i] = 0;
		}
	}
	
	grammar->V = ntypes (data);
	
	fprintf(stderr, "N = %d, K = %d, V = %d, gamma = %lg, eta = %lg\n",
		N, K, grammar->V, grammar->gamma, grammar->eta);

	for (j = 0; j < iter; j++)
	{
		z = 0; s = 0; accept = 0;
		random_shuffle (seq.begin(), seq.end(), irand);

		for (i = 0; i < N; i++)
		{
			if ((i + 1) % INTERVAL == 0) {
				fprintf(stderr, "parsing %4d/%d..\r", i+1, N);
				fflush(stderr);
			}
			n = seq[i];
			
			if (j > 0) {
				remove_counts (tree[n]);
			}
			
			newtree = parse (data[n], tree[n], slices[n]);
			newlik = polyalik (newtree);

			if (j > 0) {
				lik = polyalik (tree[n]);
				diff = newlik - lik
					+ treelik(tree[n]) - treelik(newtree);
			} else {
				diff = 0;
			}

			if (bernoulli(exp(diff)))  /* accept */
			{
				if (j > 0) delete (tree[n]);
				tree[n] = newtree;
				lik = newlik;
				accept++;
			}
			add_counts (tree[n]);
				
			z += lik;
			s += data[n].size();
			
		}
		printf("iter[%2d] : PPL = %4.2f (accept=%.2f%%)\n",
		       j + 1, exp(- z / s), (double) accept / N * 100);
	}
	    
	return true;
}

bool
PCFGS::save (char *model)
{
	save_tree (model);
	save_dict (model);
	save_grammar (model);
}

bool
PCFGS::save_dict (char *model)
{
	dict->save (model);
}

bool
PCFGS::save_grammar (char *model)
{
	grammar->save (model);
}

bool
PCFGS::save_tree (char *model)
{
	int i, N = tree.size();
	ofstream ofs (strconcat(model, ".tree"));

	if (!ofs) {
		perror("ofstream");
		return false;
	}

	for (i = 0; i < N; i++)
	{
		ofs << tree[i]->stringify(dict) << endl;
	}

	ofs.close();

	return true;
}

bool
PCFGS::read (char *file, vector<vector<int> > &data)
{
	vector<string> toks;
	string line;
	int i;
	ifstream ifs (file);

	if (!ifs) {
		perror("ifstream");
		return false;
	}

	while (getline(ifs, line, '\n'))
	{
		toks = parse_line (line);

		if (toks.size() > 0)
		{
			vector<int> row;
			
			for (i = 0; i < toks.size(); i++)
			{
				row.push_back (dict->id (toks[i]));
			}
			data.push_back (row);
		}
	}

	return true;
}
	
