/*
    pcfgt.cpp
    Bayesian PCFG with slice sampler. (Blunsom & Cohn 2010)
    $Id: pcfgt.cpp,v 1.11 2015/03/06 12:47:37 daichi Exp $

*/
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <cmath>
#include "pcfgt.h"
#include "tree.h"
#include "random.h"
#include "util.h"
#define	 INTERVAL	100

using namespace std;

typedef struct {
	vector<vector<double> > *slice;
	Grammar *grammar;
} args;

void
set_slice_func (int i, int j, int x, int y, int z, void *varg)
{
	args *arg = (args *)varg;

	(*arg->slice)[i][j] =
		RANDOM * arg->grammar->expand (x, y, z);
}

void
PCFGT::init_slice (Tree *tree, int L)
{
	int i, j;
	
	if (L > slice.size())
	{
		slice.resize (L);
		for (i = 0; i < L; i++)
		{
			slice[i].resize (L);
		}
	}
	for (i = 0; i < L; i++)
		for (j = 0; j < L; j++)
		{
			if (tree)
				slice[i][j] = betarand (alphab, betab);
			else
				slice[i][j] = 0;
		}

}

void
PCFGT::set_slice (Tree *tree, int L)
{
	args arg;
	arg.slice = &slice;
	arg.grammar = grammar;
	
	init_slice (tree, L);
	
	if (tree) {
		tree->traverse (set_slice_func, (void *)&arg);
	}
}


Tree *
PCFGT::parse (vector<int> &s, Tree *tree)
{
	set_slice (tree, s.size());
	compute_cky (s);
	return sample (s, 0, 0, s.size() - 1);
}

double
PCFGT::expand (int x, int y, int z)
{
	return grammar->expand (x, y, z);
}

double
PCFGT::expand (int x, int w)
{
	return grammar->expand (x, w);
}
	
Tree *
PCFGT::sample (vector<int> &s, int x, int i, int k)
{
	static vector<double> q;
	static vector<pair<int,pair<int,int> > > t;
	int j, a, y, z, L = k - i;
	double l, theta;

	if (i == k)
	{
		return new Tree (x, s[i]);
	}

	/* allocate */
	q.reserve (L * K * K);
	t.reserve (L * K * K);
	q.clear ();
	t.clear ();

	/* sample */
	for (j = 0; j < L; j++)
	{
		theta = slice[i][k];
			
		for (y = 1; y < K; y++)
		{
			for (z = 1; z < K; z++)
			{
				if (isinf(alpha[i][i+j][y]) ||
				    isinf(alpha[i+j+1][k][z]))
					;
				else {
					l = log (expand(x, y, z)) +
						alpha[i][i+j][y] + alpha[i+j+1][k][z];
					q.push_back (l);
					t.push_back (cons(j, cons(y, z)));
				}
			}
		}
	}

	if ((a = lmultinom(q)) < 0) {
		error("invalid draw!");
		exit(1);
		return NULL;
	} else {
		j = t[a].first;
		y = t[a].second.first;
		z = t[a].second.second;

		return new Tree (x, sample (s, y, i, i + j),
				    sample (s, z, i + j + 1, k));
	}
	
}

void
PCFGT::compute_cky (vector<int> &s)
{
	int i, j, k, n, L = s.size();
	int x, y, z;
	double l, theta;

	/* allocate */
	if (L > alpha.size()) {
		alpha.resize (L);
		for (i = 0; i < L; i++) {
			alpha[i].resize (L);
			for (j = 0; j < L; j++) {
				alpha[i][j].resize (K);
			}
		}
	}

	/* initialize */
	for (i = 0; i < L; i++) {
		for (k = 1; k < K; k++) {
			alpha[i][i][k] = log (expand(k, s[i]));
		}
	}

	/* iterate */
	
	for (n = 1; n < L; n++)
	{
		for (i = 0; i < L - n; i++)
		{
			theta = slice[i][i+n];
			
			if (n == L - 1) {
				alpha[i][i+n][0] = cky_sum (0, i, n, theta);
			}
			else {
				for (x = 1; x < K; x++)
				{
					alpha[i][i+n][x] = cky_sum (x, i, n, theta);
				}
			}
		}
	}
	
}

double
PCFGT::cky_sum (int x, int i, int n, double theta)
{
	int j, y, z;
	double p, l, lik = -inf;

	for (y = 1; y < K; y++)
	{
		for (z = 1; z < K; z++)
		{
			p = expand (x, y, z);

			if (p > theta)
			{
				l = 0;
				for (j = 0; j < n; j++)
				{
					l = logsumexp (l, alpha[i][i+j][y] +
						          alpha[i+j+1][i+n][z], (j == 0));
				}
				lik = logsumexp (lik, l - lbetad(theta, alphab, betab));
			} else {
				lik = logsumexp (lik, -inf);
			}
		}
	}

	return lik;
}

bool
PCFGT::add_counts (Tree *tree)
{
	return grammar->add_counts (tree);
}

bool
PCFGT::remove_counts (Tree *tree)
{
	return grammar->remove_counts (tree);
}

double
PCFGT::treelik (Tree *tree)
{
	return grammar->treelik (tree);
}
    
double
PCFGT::polyalik (Tree *tree)
{
	return grammar->polyalik (tree);
}

bool
PCFGT::train (vector<vector<int> > &data, int iter)
{
	int i, j, n, N = data.size();
	double s, z, diff, lik, newlik;
	Tree *newtree;
	int accept;

	tree.resize (N);
	
	for (n = 0; n < N; n++)
	{
		tree[n] = NULL;
	}
	
	grammar->V = ntypes (data);
	
	fprintf(stderr, "N = %d, K = %d, V = %d, gamma = %lg, eta = %lg\n",
		N, K, grammar->V, grammar->gamma, grammar->eta);

	for (j = 0; j < iter; j++)
	{
		z = 0; s = 0; accept = 0;
		
		for (n = 0; n < N; n++)
		{
			if ((n + 1) % INTERVAL == 0)
			{
				fprintf(stderr, "parsing %4d/%d..\r", n+1, N);
				fflush(stderr);
			}
			
			if (j > 0) {
				remove_counts (tree[n]);
			}
			
			newtree = parse (data[n], tree[n]);
			newlik = polyalik (newtree);

			if (j > 0) {
				lik = polyalik (tree[n]);
				diff = newlik - lik
					+ treelik(tree[n]) - treelik(newtree);
			} else {
				diff = 0;
			}

			if (bernoulli(exp(diff)))
			{
				if (j > 0) delete (tree[n]);
				tree[n] = newtree;
				lik = newlik;
				accept++;
			}
				
			add_counts (tree[n]);
				
			z += lik;
			s += data[n].size();
			
		}
		printf("iter[%2d] : PPL = %4.2f (accept=%.2f%%)\n",
		       j + 1, exp(- z / s), (double) accept / N * 100);
	}
	    
	return true;
}

bool
PCFGT::save (char *model)
{
	save_tree (model);
	save_dict (model);
	save_grammar (model);
}

bool
PCFGT::save_dict (char *model)
{
	dict->save (model);
}

bool
PCFGT::save_grammar (char *model)
{
	grammar->save (model);
}

bool
PCFGT::save_tree (char *model)
{
	int i, N = tree.size();
	ofstream ofs (strconcat(model, ".tree"));

	if (!ofs) {
		perror("ofstream");
		return false;
	}

	for (i = 0; i < N; i++)
	{
		ofs << tree[i]->stringify(dict) << endl;
	}

	ofs.close();

	return true;
}

bool
PCFGT::read (char *file, vector<vector<int> > &data)
{
	vector<string> toks;
	string line;
	int i, w;
	ifstream ifs (file);

	if (!ifs) {
		perror("ifstream");
		return false;
	}

	while (getline(ifs, line, '\n'))
	{
		toks = parse_line (line);

		if (toks.size() > 0)
		{
			vector<int> row;
			
			for (i = 0; i < toks.size(); i++)
			{
				row.push_back (dict->id (toks[i]));
			}
			data.push_back (row);
		}
	}

	return true;
}
	
